const std = @import("std");
const builtin = @import("builtin");
const Clipboard = @import("clipboard.zig");
const w = std.os.windows;

extern "user32" fn MessageBoxA(hWnd: ?w.HANDLE, lpText: ?w.LPCSTR, lpCaption: ?w.LPCSTR, uType: w.UINT) callconv(w.WINAPI) c_int;

const WINDOW_EX_STYLE = enum(u32) {
    DLGMODALFRAME = 1,
    NOPARENTNOTIFY = 4,
    TOPMOST = 8,
    ACCEPTFILES = 16,
    TRANSPARENT = 32,
    MDICHILD = 64,
    TOOLWINDOW = 128,
    WINDOWEDGE = 256,
    CLIENTEDGE = 512,
    CONTEXTHELP = 1024,
    RIGHT = 4096,
    LEFT = 0,
    RTLREADING = 8192,
    // LTRREADING = 0, this enum value conflicts with LEFT
    LEFTSCROLLBAR = 16384,
    // RIGHTSCROLLBAR = 0, this enum value conflicts with LEFT
    CONTROLPARENT = 65536,
    STATICEDGE = 131072,
    APPWINDOW = 262144,
    OVERLAPPEDWINDOW = 768,
    PALETTEWINDOW = 392,
    LAYERED = 524288,
    NOINHERITLAYOUT = 1048576,
    NOREDIRECTIONBITMAP = 2097152,
    LAYOUTRTL = 4194304,
    COMPOSITED = 33554432,
    NOACTIVATE = 134217728,
    _,
    pub fn initFlags(o: struct {
        DLGMODALFRAME: u1 = 0,
        NOPARENTNOTIFY: u1 = 0,
        TOPMOST: u1 = 0,
        ACCEPTFILES: u1 = 0,
        TRANSPARENT: u1 = 0,
        MDICHILD: u1 = 0,
        TOOLWINDOW: u1 = 0,
        WINDOWEDGE: u1 = 0,
        CLIENTEDGE: u1 = 0,
        CONTEXTHELP: u1 = 0,
        RIGHT: u1 = 0,
        LEFT: u1 = 0,
        RTLREADING: u1 = 0,
        LEFTSCROLLBAR: u1 = 0,
        CONTROLPARENT: u1 = 0,
        STATICEDGE: u1 = 0,
        APPWINDOW: u1 = 0,
        OVERLAPPEDWINDOW: u1 = 0,
        PALETTEWINDOW: u1 = 0,
        LAYERED: u1 = 0,
        NOINHERITLAYOUT: u1 = 0,
        NOREDIRECTIONBITMAP: u1 = 0,
        LAYOUTRTL: u1 = 0,
        COMPOSITED: u1 = 0,
        NOACTIVATE: u1 = 0,
    }) WINDOW_EX_STYLE {
        return @intToEnum(WINDOW_EX_STYLE, (if (o.DLGMODALFRAME == 1) @enumToInt(WINDOW_EX_STYLE.DLGMODALFRAME) else 0) | (if (o.NOPARENTNOTIFY == 1) @enumToInt(WINDOW_EX_STYLE.NOPARENTNOTIFY) else 0) | (if (o.TOPMOST == 1) @enumToInt(WINDOW_EX_STYLE.TOPMOST) else 0) | (if (o.ACCEPTFILES == 1) @enumToInt(WINDOW_EX_STYLE.ACCEPTFILES) else 0) | (if (o.TRANSPARENT == 1) @enumToInt(WINDOW_EX_STYLE.TRANSPARENT) else 0) | (if (o.MDICHILD == 1) @enumToInt(WINDOW_EX_STYLE.MDICHILD) else 0) | (if (o.TOOLWINDOW == 1) @enumToInt(WINDOW_EX_STYLE.TOOLWINDOW) else 0) | (if (o.WINDOWEDGE == 1) @enumToInt(WINDOW_EX_STYLE.WINDOWEDGE) else 0) | (if (o.CLIENTEDGE == 1) @enumToInt(WINDOW_EX_STYLE.CLIENTEDGE) else 0) | (if (o.CONTEXTHELP == 1) @enumToInt(WINDOW_EX_STYLE.CONTEXTHELP) else 0) | (if (o.RIGHT == 1) @enumToInt(WINDOW_EX_STYLE.RIGHT) else 0) | (if (o.LEFT == 1) @enumToInt(WINDOW_EX_STYLE.LEFT) else 0) | (if (o.RTLREADING == 1) @enumToInt(WINDOW_EX_STYLE.RTLREADING) else 0) | (if (o.LEFTSCROLLBAR == 1) @enumToInt(WINDOW_EX_STYLE.LEFTSCROLLBAR) else 0) | (if (o.CONTROLPARENT == 1) @enumToInt(WINDOW_EX_STYLE.CONTROLPARENT) else 0) | (if (o.STATICEDGE == 1) @enumToInt(WINDOW_EX_STYLE.STATICEDGE) else 0) | (if (o.APPWINDOW == 1) @enumToInt(WINDOW_EX_STYLE.APPWINDOW) else 0) | (if (o.OVERLAPPEDWINDOW == 1) @enumToInt(WINDOW_EX_STYLE.OVERLAPPEDWINDOW) else 0) | (if (o.PALETTEWINDOW == 1) @enumToInt(WINDOW_EX_STYLE.PALETTEWINDOW) else 0) | (if (o.LAYERED == 1) @enumToInt(WINDOW_EX_STYLE.LAYERED) else 0) | (if (o.NOINHERITLAYOUT == 1) @enumToInt(WINDOW_EX_STYLE.NOINHERITLAYOUT) else 0) | (if (o.NOREDIRECTIONBITMAP == 1) @enumToInt(WINDOW_EX_STYLE.NOREDIRECTIONBITMAP) else 0) | (if (o.LAYOUTRTL == 1) @enumToInt(WINDOW_EX_STYLE.LAYOUTRTL) else 0) | (if (o.COMPOSITED == 1) @enumToInt(WINDOW_EX_STYLE.COMPOSITED) else 0) | (if (o.NOACTIVATE == 1) @enumToInt(WINDOW_EX_STYLE.NOACTIVATE) else 0));
    }
};

extern "USER32" fn CreateWindowExA(
    dwExStyle: WINDOW_EX_STYLE,
    lpClassName: ?[*:0]const u8,
    lpWindowName: ?[*:0]const u8,
    dwStyle: WINDOW_STYLE,
    X: i32,
    Y: i32,
    nWidth: i32,
    nHeight: i32,
    hWndParent: ?w.HWND,
    hMenu: ?w.HMENU,
    hInstance: ?w.HINSTANCE,
    lpParam: ?*anyopaque,
) callconv(w.WINAPI) ?w.HWND;

const WNDPROC = fn (
    param0: w.HWND,
    param1: u32,
    param2: w.WPARAM,
    param3: w.LPARAM,
) callconv(w.WINAPI) w.LRESULT;

const WNDCLASS_STYLES = enum(u32) {
    VREDRAW = 1,
    HREDRAW = 2,
    DBLCLKS = 8,
    OWNDC = 32,
    CLASSDC = 64,
    PARENTDC = 128,
    NOCLOSE = 512,
    SAVEBITS = 2048,
    BYTEALIGNCLIENT = 4096,
    BYTEALIGNWINDOW = 8192,
    GLOBALCLASS = 16384,
    IME = 65536,
    DROPSHADOW = 131072,
    _,
    pub fn initFlags(o: struct {
        VREDRAW: u1 = 0,
        HREDRAW: u1 = 0,
        DBLCLKS: u1 = 0,
        OWNDC: u1 = 0,
        CLASSDC: u1 = 0,
        PARENTDC: u1 = 0,
        NOCLOSE: u1 = 0,
        SAVEBITS: u1 = 0,
        BYTEALIGNCLIENT: u1 = 0,
        BYTEALIGNWINDOW: u1 = 0,
        GLOBALCLASS: u1 = 0,
        IME: u1 = 0,
        DROPSHADOW: u1 = 0,
    }) WNDCLASS_STYLES {
        return @intToEnum(WNDCLASS_STYLES, (if (o.VREDRAW == 1) @enumToInt(WNDCLASS_STYLES.VREDRAW) else 0) | (if (o.HREDRAW == 1) @enumToInt(WNDCLASS_STYLES.HREDRAW) else 0) | (if (o.DBLCLKS == 1) @enumToInt(WNDCLASS_STYLES.DBLCLKS) else 0) | (if (o.OWNDC == 1) @enumToInt(WNDCLASS_STYLES.OWNDC) else 0) | (if (o.CLASSDC == 1) @enumToInt(WNDCLASS_STYLES.CLASSDC) else 0) | (if (o.PARENTDC == 1) @enumToInt(WNDCLASS_STYLES.PARENTDC) else 0) | (if (o.NOCLOSE == 1) @enumToInt(WNDCLASS_STYLES.NOCLOSE) else 0) | (if (o.SAVEBITS == 1) @enumToInt(WNDCLASS_STYLES.SAVEBITS) else 0) | (if (o.BYTEALIGNCLIENT == 1) @enumToInt(WNDCLASS_STYLES.BYTEALIGNCLIENT) else 0) | (if (o.BYTEALIGNWINDOW == 1) @enumToInt(WNDCLASS_STYLES.BYTEALIGNWINDOW) else 0) | (if (o.GLOBALCLASS == 1) @enumToInt(WNDCLASS_STYLES.GLOBALCLASS) else 0) | (if (o.IME == 1) @enumToInt(WNDCLASS_STYLES.IME) else 0) | (if (o.DROPSHADOW == 1) @enumToInt(WNDCLASS_STYLES.DROPSHADOW) else 0));
    }
};
const WINDOW_STYLE = enum(u32) {
    OVERLAPPED = 0,
    POPUP = 2147483648,
    CHILD = 1073741824,
    MINIMIZE = 536870912,
    VISIBLE = 268435456,
    DISABLED = 134217728,
    CLIPSIBLINGS = 67108864,
    CLIPCHILDREN = 33554432,
    MAXIMIZE = 16777216,
    CAPTION = 12582912,
    BORDER = 8388608,
    DLGFRAME = 4194304,
    VSCROLL = 2097152,
    HSCROLL = 1048576,
    SYSMENU = 524288,
    THICKFRAME = 262144,
    GROUP = 131072,
    TABSTOP = 65536,
    // MINIMIZEBOX = 131072, this enum value conflicts with GROUP
    // MAXIMIZEBOX = 65536, this enum value conflicts with TABSTOP
    // TILED = 0, this enum value conflicts with OVERLAPPED
    // ICONIC = 536870912, this enum value conflicts with MINIMIZE
    // SIZEBOX = 262144, this enum value conflicts with THICKFRAME
    TILEDWINDOW = 13565952,
    // OVERLAPPEDWINDOW = 13565952, this enum value conflicts with TILEDWINDOW
    POPUPWINDOW = 2156396544,
    // CHILDWINDOW = 1073741824, this enum value conflicts with CHILD
    ACTIVECAPTION = 1,
    _,
    pub fn initFlags(o: struct {
        OVERLAPPED: u1 = 0,
        POPUP: u1 = 0,
        CHILD: u1 = 0,
        MINIMIZE: u1 = 0,
        VISIBLE: u1 = 0,
        DISABLED: u1 = 0,
        CLIPSIBLINGS: u1 = 0,
        CLIPCHILDREN: u1 = 0,
        MAXIMIZE: u1 = 0,
        CAPTION: u1 = 0,
        BORDER: u1 = 0,
        DLGFRAME: u1 = 0,
        VSCROLL: u1 = 0,
        HSCROLL: u1 = 0,
        SYSMENU: u1 = 0,
        THICKFRAME: u1 = 0,
        GROUP: u1 = 0,
        TABSTOP: u1 = 0,
        TILEDWINDOW: u1 = 0,
        POPUPWINDOW: u1 = 0,
        ACTIVECAPTION: u1 = 0,
    }) WINDOW_STYLE {
        return @intToEnum(WINDOW_STYLE, (if (o.OVERLAPPED == 1) @enumToInt(WINDOW_STYLE.OVERLAPPED) else 0) | (if (o.POPUP == 1) @enumToInt(WINDOW_STYLE.POPUP) else 0) | (if (o.CHILD == 1) @enumToInt(WINDOW_STYLE.CHILD) else 0) | (if (o.MINIMIZE == 1) @enumToInt(WINDOW_STYLE.MINIMIZE) else 0) | (if (o.VISIBLE == 1) @enumToInt(WINDOW_STYLE.VISIBLE) else 0) | (if (o.DISABLED == 1) @enumToInt(WINDOW_STYLE.DISABLED) else 0) | (if (o.CLIPSIBLINGS == 1) @enumToInt(WINDOW_STYLE.CLIPSIBLINGS) else 0) | (if (o.CLIPCHILDREN == 1) @enumToInt(WINDOW_STYLE.CLIPCHILDREN) else 0) | (if (o.MAXIMIZE == 1) @enumToInt(WINDOW_STYLE.MAXIMIZE) else 0) | (if (o.CAPTION == 1) @enumToInt(WINDOW_STYLE.CAPTION) else 0) | (if (o.BORDER == 1) @enumToInt(WINDOW_STYLE.BORDER) else 0) | (if (o.DLGFRAME == 1) @enumToInt(WINDOW_STYLE.DLGFRAME) else 0) | (if (o.VSCROLL == 1) @enumToInt(WINDOW_STYLE.VSCROLL) else 0) | (if (o.HSCROLL == 1) @enumToInt(WINDOW_STYLE.HSCROLL) else 0) | (if (o.SYSMENU == 1) @enumToInt(WINDOW_STYLE.SYSMENU) else 0) | (if (o.THICKFRAME == 1) @enumToInt(WINDOW_STYLE.THICKFRAME) else 0) | (if (o.GROUP == 1) @enumToInt(WINDOW_STYLE.GROUP) else 0) | (if (o.TABSTOP == 1) @enumToInt(WINDOW_STYLE.TABSTOP) else 0) | (if (o.TILEDWINDOW == 1) @enumToInt(WINDOW_STYLE.TILEDWINDOW) else 0) | (if (o.POPUPWINDOW == 1) @enumToInt(WINDOW_STYLE.POPUPWINDOW) else 0) | (if (o.ACTIVECAPTION == 1) @enumToInt(WINDOW_STYLE.ACTIVECAPTION) else 0));
    }
};

const WNDCLASSA = extern struct {
    style: WNDCLASS_STYLES,
    lpfnWndProc: ?WNDPROC,
    cbClsExtra: i32,
    cbWndExtra: i32,
    hInstance: ?w.HINSTANCE,
    hIcon: ?w.HICON,
    hCursor: ?w.HCURSOR,
    hbrBackground: ?w.HBRUSH,
    lpszMenuName: ?[*:0]const u8,
    lpszClassName: ?[*:0]const u8,
};

extern "USER32" fn RegisterClassA(
    lpWndClass: ?*const WNDCLASSA,
) callconv(w.WINAPI) u16;

extern "USER32" fn DestroyWindow(
    hWnd: ?w.HWND,
) callconv(w.WINAPI) w.BOOL;

const WS_OVERLAPPEDWINDOW = WINDOW_STYLE.TILEDWINDOW;
const CW_USEDEFAULT = @as(i32, -2147483648);
const BOOL = i32;

const MSG = extern struct {
    hwnd: ?w.HWND,
    message: u32,
    wParam: w.WPARAM,
    lParam: w.LPARAM,
    time: u32,
    pt: w.POINT,
};
const SHOW_WINDOW_CMD = enum(u32) {
    FORCEMINIMIZE = 11,
    HIDE = 0,
    MAXIMIZE = 3,
    MINIMIZE = 6,
    RESTORE = 9,
    SHOW = 5,
    SHOWDEFAULT = 10,
    // SHOWMAXIMIZED = 3, this enum value conflicts with MAXIMIZE
    SHOWMINIMIZED = 2,
    SHOWMINNOACTIVE = 7,
    SHOWNA = 8,
    SHOWNOACTIVATE = 4,
    SHOWNORMAL = 1,
    // NORMAL = 1, this enum value conflicts with SHOWNORMAL
    // MAX = 11, this enum value conflicts with FORCEMINIMIZE
    // PARENTCLOSING = 1, this enum value conflicts with SHOWNORMAL
    // OTHERZOOM = 2, this enum value conflicts with SHOWMINIMIZED
    // PARENTOPENING = 3, this enum value conflicts with MAXIMIZE
    // OTHERUNZOOM = 4, this enum value conflicts with SHOWNOACTIVATE
    // SCROLLCHILDREN = 1, this enum value conflicts with SHOWNORMAL
    // INVALIDATE = 2, this enum value conflicts with SHOWMINIMIZED
    // ERASE = 4, this enum value conflicts with SHOWNOACTIVATE
    SMOOTHSCROLL = 16,
    _,
    pub fn initFlags(o: struct {
        FORCEMINIMIZE: u1 = 0,
        HIDE: u1 = 0,
        MAXIMIZE: u1 = 0,
        MINIMIZE: u1 = 0,
        RESTORE: u1 = 0,
        SHOW: u1 = 0,
        SHOWDEFAULT: u1 = 0,
        SHOWMINIMIZED: u1 = 0,
        SHOWMINNOACTIVE: u1 = 0,
        SHOWNA: u1 = 0,
        SHOWNOACTIVATE: u1 = 0,
        SHOWNORMAL: u1 = 0,
        SMOOTHSCROLL: u1 = 0,
    }) SHOW_WINDOW_CMD {
        return @intToEnum(SHOW_WINDOW_CMD, (if (o.FORCEMINIMIZE == 1) @enumToInt(SHOW_WINDOW_CMD.FORCEMINIMIZE) else 0) | (if (o.HIDE == 1) @enumToInt(SHOW_WINDOW_CMD.HIDE) else 0) | (if (o.MAXIMIZE == 1) @enumToInt(SHOW_WINDOW_CMD.MAXIMIZE) else 0) | (if (o.MINIMIZE == 1) @enumToInt(SHOW_WINDOW_CMD.MINIMIZE) else 0) | (if (o.RESTORE == 1) @enumToInt(SHOW_WINDOW_CMD.RESTORE) else 0) | (if (o.SHOW == 1) @enumToInt(SHOW_WINDOW_CMD.SHOW) else 0) | (if (o.SHOWDEFAULT == 1) @enumToInt(SHOW_WINDOW_CMD.SHOWDEFAULT) else 0) | (if (o.SHOWMINIMIZED == 1) @enumToInt(SHOW_WINDOW_CMD.SHOWMINIMIZED) else 0) | (if (o.SHOWMINNOACTIVE == 1) @enumToInt(SHOW_WINDOW_CMD.SHOWMINNOACTIVE) else 0) | (if (o.SHOWNA == 1) @enumToInt(SHOW_WINDOW_CMD.SHOWNA) else 0) | (if (o.SHOWNOACTIVATE == 1) @enumToInt(SHOW_WINDOW_CMD.SHOWNOACTIVATE) else 0) | (if (o.SHOWNORMAL == 1) @enumToInt(SHOW_WINDOW_CMD.SHOWNORMAL) else 0) | (if (o.SMOOTHSCROLL == 1) @enumToInt(SHOW_WINDOW_CMD.SMOOTHSCROLL) else 0));
    }
};

extern "USER32" fn ShowWindow(
    hWnd: ?w.HWND,
    nCmdShow: SHOW_WINDOW_CMD,
) callconv(w.WINAPI) BOOL;

extern "USER32" fn GetMessageA(
    lpMsg: ?*MSG,
    hWnd: ?w.HWND,
    wMsgFilterMin: u32,
    wMsgFilterMax: u32,
) callconv(w.WINAPI) BOOL;

extern "USER32" fn TranslateMessage(
    lpMsg: ?*const MSG,
) callconv(w.WINAPI) BOOL;

const LRESULT = isize;

extern "USER32" fn DispatchMessageA(
    lpMsg: ?*const MSG,
) callconv(w.WINAPI) LRESULT;

const WM_NULL = @as(u32, 0);
const WM_CREATE = @as(u32, 1);
const WM_DESTROY = @as(u32, 2);
const WM_MOVE = @as(u32, 3);
const WM_SIZE = @as(u32, 5);
const WM_ACTIVATE = @as(u32, 6);
const WA_INACTIVE = @as(u32, 0);
const WA_ACTIVE = @as(u32, 1);
const WA_CLICKACTIVE = @as(u32, 2);
const WM_SETFOCUS = @as(u32, 7);
const WM_KILLFOCUS = @as(u32, 8);
const WM_ENABLE = @as(u32, 10);
const WM_SETREDRAW = @as(u32, 11);
const WM_SETTEXT = @as(u32, 12);
const WM_GETTEXT = @as(u32, 13);
const WM_GETTEXTLENGTH = @as(u32, 14);
const WM_PAINT = @as(u32, 15);
const WM_CLOSE = @as(u32, 16);
const WM_QUERYENDSESSION = @as(u32, 17);
const WM_QUERYOPEN = @as(u32, 19);
const WM_ENDSESSION = @as(u32, 22);
const WM_QUIT = @as(u32, 18);
const WM_ERASEBKGND = @as(u32, 20);
const WM_SYSCOLORCHANGE = @as(u32, 21);
const WM_SHOWWINDOW = @as(u32, 24);
const WM_WININICHANGE = @as(u32, 26);
const WM_SETTINGCHANGE = @as(u32, 26);
const WM_DEVMODECHANGE = @as(u32, 27);
const WM_ACTIVATEAPP = @as(u32, 28);
const WM_FONTCHANGE = @as(u32, 29);
const WM_TIMECHANGE = @as(u32, 30);
const WM_CANCELMODE = @as(u32, 31);
const WM_SETCURSOR = @as(u32, 32);
const WM_MOUSEACTIVATE = @as(u32, 33);
const WM_CHILDACTIVATE = @as(u32, 34);
const WM_QUEUESYNC = @as(u32, 35);
const WM_GETMINMAXINFO = @as(u32, 36);
const WM_PAINTICON = @as(u32, 38);
const WM_ICONERASEBKGND = @as(u32, 39);
const WM_NEXTDLGCTL = @as(u32, 40);
const WM_SPOOLERSTATUS = @as(u32, 42);
const WM_DRAWITEM = @as(u32, 43);
const WM_MEASUREITEM = @as(u32, 44);
const WM_DELETEITEM = @as(u32, 45);
const WM_VKEYTOITEM = @as(u32, 46);
const WM_CHARTOITEM = @as(u32, 47);
const WM_SETFONT = @as(u32, 48);
const WM_GETFONT = @as(u32, 49);
const WM_SETHOTKEY = @as(u32, 50);
const WM_GETHOTKEY = @as(u32, 51);
const WM_QUERYDRAGICON = @as(u32, 55);
const WM_COMPAREITEM = @as(u32, 57);
const WM_GETOBJECT = @as(u32, 61);
const WM_COMPACTING = @as(u32, 65);
const WM_COMMNOTIFY = @as(u32, 68);
const WM_WINDOWPOSCHANGING = @as(u32, 70);
const WM_WINDOWPOSCHANGED = @as(u32, 71);
const WM_POWER = @as(u32, 72);

const WM_COMMAND = @as(u32, 273);
const WM_INITMENU = @as(u32, 278);
const WM_INITMENUPOPUP = @as(u32, 279);

const WM_DRAWCLIPBOARD = @as(u32, 776);
const WM_SIZECLIPBOARD = @as(u32, 779);
const WM_CHANGECBCHAIN = @as(u32, 781);
const WM_CLIPBOARDUPDATE = @as(u32, 797);

const WM_USER = @as(u32, 1024);
// USER constants here - not part of Windows headers
const WM_TRAY = @as(u32, WM_USER + 1);

const RECT = extern struct {
    left: i32,
    top: i32,
    right: i32,
    bottom: i32,
};

extern "USER32" fn GetClientRect(
    hWnd: ?w.HWND,
    lpRect: ?*RECT,
) callconv(w.WINAPI) BOOL;

extern "USER32" fn SendMessageA(
    hWnd: ?w.HWND,
    Msg: u32,
    wParam: w.WPARAM,
    lParam: w.LPARAM,
) callconv(w.WINAPI) LRESULT;

extern "USER32" fn DefWindowProcA(
    hWnd: ?w.HWND,
    Msg: u32,
    wParam: w.WPARAM,
    lParam: w.LPARAM,
) callconv(w.WINAPI) LRESULT;

// GDI
const HDC = *opaque {};

pub const PAINTSTRUCT = extern struct {
    hdc: ?HDC,
    fErase: BOOL,
    rcPaint: RECT,
    fRestore: BOOL,
    fIncUpdate: BOOL,
    rgbReserved: [32]u8,
};

extern "USER32" fn BeginPaint(
    hWnd: ?w.HWND,
    lpPaint: ?*PAINTSTRUCT,
) callconv(w.WINAPI) ?HDC;

extern "USER32" fn EndPaint(
    hWnd: ?w.HWND,
    lpPaint: ?*const PAINTSTRUCT,
) callconv(w.WINAPI) BOOL;

const DRAW_TEXT_FORMAT = enum(u32) {
    BOTTOM = 8,
    CALCRECT = 1024,
    CENTER = 1,
    EDITCONTROL = 8192,
    END_ELLIPSIS = 32768,
    EXPANDTABS = 64,
    EXTERNALLEADING = 512,
    HIDEPREFIX = 1048576,
    INTERNAL = 4096,
    LEFT = 0,
    MODIFYSTRING = 65536,
    NOCLIP = 256,
    NOFULLWIDTHCHARBREAK = 524288,
    NOPREFIX = 2048,
    PATH_ELLIPSIS = 16384,
    PREFIXONLY = 2097152,
    RIGHT = 2,
    RTLREADING = 131072,
    SINGLELINE = 32,
    TABSTOP = 128,
    // TOP = 0, this enum value conflicts with LEFT
    VCENTER = 4,
    WORDBREAK = 16,
    WORD_ELLIPSIS = 262144,
    _,
    pub fn initFlags(o: struct {
        BOTTOM: u1 = 0,
        CALCRECT: u1 = 0,
        CENTER: u1 = 0,
        EDITCONTROL: u1 = 0,
        END_ELLIPSIS: u1 = 0,
        EXPANDTABS: u1 = 0,
        EXTERNALLEADING: u1 = 0,
        HIDEPREFIX: u1 = 0,
        INTERNAL: u1 = 0,
        LEFT: u1 = 0,
        MODIFYSTRING: u1 = 0,
        NOCLIP: u1 = 0,
        NOFULLWIDTHCHARBREAK: u1 = 0,
        NOPREFIX: u1 = 0,
        PATH_ELLIPSIS: u1 = 0,
        PREFIXONLY: u1 = 0,
        RIGHT: u1 = 0,
        RTLREADING: u1 = 0,
        SINGLELINE: u1 = 0,
        TABSTOP: u1 = 0,
        VCENTER: u1 = 0,
        WORDBREAK: u1 = 0,
        WORD_ELLIPSIS: u1 = 0,
    }) DRAW_TEXT_FORMAT {
        return @intToEnum(DRAW_TEXT_FORMAT, (if (o.BOTTOM == 1) @enumToInt(DRAW_TEXT_FORMAT.BOTTOM) else 0) | (if (o.CALCRECT == 1) @enumToInt(DRAW_TEXT_FORMAT.CALCRECT) else 0) | (if (o.CENTER == 1) @enumToInt(DRAW_TEXT_FORMAT.CENTER) else 0) | (if (o.EDITCONTROL == 1) @enumToInt(DRAW_TEXT_FORMAT.EDITCONTROL) else 0) | (if (o.END_ELLIPSIS == 1) @enumToInt(DRAW_TEXT_FORMAT.END_ELLIPSIS) else 0) | (if (o.EXPANDTABS == 1) @enumToInt(DRAW_TEXT_FORMAT.EXPANDTABS) else 0) | (if (o.EXTERNALLEADING == 1) @enumToInt(DRAW_TEXT_FORMAT.EXTERNALLEADING) else 0) | (if (o.HIDEPREFIX == 1) @enumToInt(DRAW_TEXT_FORMAT.HIDEPREFIX) else 0) | (if (o.INTERNAL == 1) @enumToInt(DRAW_TEXT_FORMAT.INTERNAL) else 0) | (if (o.LEFT == 1) @enumToInt(DRAW_TEXT_FORMAT.LEFT) else 0) | (if (o.MODIFYSTRING == 1) @enumToInt(DRAW_TEXT_FORMAT.MODIFYSTRING) else 0) | (if (o.NOCLIP == 1) @enumToInt(DRAW_TEXT_FORMAT.NOCLIP) else 0) | (if (o.NOFULLWIDTHCHARBREAK == 1) @enumToInt(DRAW_TEXT_FORMAT.NOFULLWIDTHCHARBREAK) else 0) | (if (o.NOPREFIX == 1) @enumToInt(DRAW_TEXT_FORMAT.NOPREFIX) else 0) | (if (o.PATH_ELLIPSIS == 1) @enumToInt(DRAW_TEXT_FORMAT.PATH_ELLIPSIS) else 0) | (if (o.PREFIXONLY == 1) @enumToInt(DRAW_TEXT_FORMAT.PREFIXONLY) else 0) | (if (o.RIGHT == 1) @enumToInt(DRAW_TEXT_FORMAT.RIGHT) else 0) | (if (o.RTLREADING == 1) @enumToInt(DRAW_TEXT_FORMAT.RTLREADING) else 0) | (if (o.SINGLELINE == 1) @enumToInt(DRAW_TEXT_FORMAT.SINGLELINE) else 0) | (if (o.TABSTOP == 1) @enumToInt(DRAW_TEXT_FORMAT.TABSTOP) else 0) | (if (o.VCENTER == 1) @enumToInt(DRAW_TEXT_FORMAT.VCENTER) else 0) | (if (o.WORDBREAK == 1) @enumToInt(DRAW_TEXT_FORMAT.WORDBREAK) else 0) | (if (o.WORD_ELLIPSIS == 1) @enumToInt(DRAW_TEXT_FORMAT.WORD_ELLIPSIS) else 0));
    }
};

const DT_BOTTOM = DRAW_TEXT_FORMAT.BOTTOM;
const DT_CALCRECT = DRAW_TEXT_FORMAT.CALCRECT;
const DT_CENTER = DRAW_TEXT_FORMAT.CENTER;
const DT_EDITCONTROL = DRAW_TEXT_FORMAT.EDITCONTROL;
const DT_END_ELLIPSIS = DRAW_TEXT_FORMAT.END_ELLIPSIS;
const DT_EXPANDTABS = DRAW_TEXT_FORMAT.EXPANDTABS;
const DT_EXTERNALLEADING = DRAW_TEXT_FORMAT.EXTERNALLEADING;
const DT_HIDEPREFIX = DRAW_TEXT_FORMAT.HIDEPREFIX;
const DT_INTERNAL = DRAW_TEXT_FORMAT.INTERNAL;
const DT_LEFT = DRAW_TEXT_FORMAT.LEFT;
const DT_MODIFYSTRING = DRAW_TEXT_FORMAT.MODIFYSTRING;
const DT_NOCLIP = DRAW_TEXT_FORMAT.NOCLIP;
const DT_NOFULLWIDTHCHARBREAK = DRAW_TEXT_FORMAT.NOFULLWIDTHCHARBREAK;
const DT_NOPREFIX = DRAW_TEXT_FORMAT.NOPREFIX;
const DT_PATH_ELLIPSIS = DRAW_TEXT_FORMAT.PATH_ELLIPSIS;
const DT_PREFIXONLY = DRAW_TEXT_FORMAT.PREFIXONLY;
const DT_RIGHT = DRAW_TEXT_FORMAT.RIGHT;
const DT_RTLREADING = DRAW_TEXT_FORMAT.RTLREADING;
const DT_SINGLELINE = DRAW_TEXT_FORMAT.SINGLELINE;
const DT_TABSTOP = DRAW_TEXT_FORMAT.TABSTOP;
const DT_TOP = DRAW_TEXT_FORMAT.LEFT;
const DT_VCENTER = DRAW_TEXT_FORMAT.VCENTER;
const DT_WORDBREAK = DRAW_TEXT_FORMAT.WORDBREAK;
const DT_WORD_ELLIPSIS = DRAW_TEXT_FORMAT.WORD_ELLIPSIS;

extern "USER32" fn DrawTextA(
    hdc: ?HDC,
    lpchText: [*:0]const u8,
    cchText: i32,
    lprc: ?*RECT,
    format: DRAW_TEXT_FORMAT,
) callconv(w.WINAPI) i32;

extern "USER32" fn InvalidateRect(
    hWnd: ?w.HWND,
    lpRect: ?*const RECT,
    bErase: w.BOOL,
) callconv(w.WINAPI) BOOL;

extern "USER32" fn UpdateWindow(
    hWnd: ?w.HWND,
) callconv(w.WINAPI) BOOL;

const REDRAW_WINDOW_FLAGS = enum(u32) {
    INVALIDATE = 1,
    INTERNALPAINT = 2,
    ERASE = 4,
    VALIDATE = 8,
    NOINTERNALPAINT = 16,
    NOERASE = 32,
    NOCHILDREN = 64,
    ALLCHILDREN = 128,
    UPDATENOW = 256,
    ERASENOW = 512,
    FRAME = 1024,
    NOFRAME = 2048,
    _,
    pub fn initFlags(o: struct {
        INVALIDATE: u1 = 0,
        INTERNALPAINT: u1 = 0,
        ERASE: u1 = 0,
        VALIDATE: u1 = 0,
        NOINTERNALPAINT: u1 = 0,
        NOERASE: u1 = 0,
        NOCHILDREN: u1 = 0,
        ALLCHILDREN: u1 = 0,
        UPDATENOW: u1 = 0,
        ERASENOW: u1 = 0,
        FRAME: u1 = 0,
        NOFRAME: u1 = 0,
    }) REDRAW_WINDOW_FLAGS {
        return @intToEnum(REDRAW_WINDOW_FLAGS, (if (o.INVALIDATE == 1) @enumToInt(REDRAW_WINDOW_FLAGS.INVALIDATE) else 0) | (if (o.INTERNALPAINT == 1) @enumToInt(REDRAW_WINDOW_FLAGS.INTERNALPAINT) else 0) | (if (o.ERASE == 1) @enumToInt(REDRAW_WINDOW_FLAGS.ERASE) else 0) | (if (o.VALIDATE == 1) @enumToInt(REDRAW_WINDOW_FLAGS.VALIDATE) else 0) | (if (o.NOINTERNALPAINT == 1) @enumToInt(REDRAW_WINDOW_FLAGS.NOINTERNALPAINT) else 0) | (if (o.NOERASE == 1) @enumToInt(REDRAW_WINDOW_FLAGS.NOERASE) else 0) | (if (o.NOCHILDREN == 1) @enumToInt(REDRAW_WINDOW_FLAGS.NOCHILDREN) else 0) | (if (o.ALLCHILDREN == 1) @enumToInt(REDRAW_WINDOW_FLAGS.ALLCHILDREN) else 0) | (if (o.UPDATENOW == 1) @enumToInt(REDRAW_WINDOW_FLAGS.UPDATENOW) else 0) | (if (o.ERASENOW == 1) @enumToInt(REDRAW_WINDOW_FLAGS.ERASENOW) else 0) | (if (o.FRAME == 1) @enumToInt(REDRAW_WINDOW_FLAGS.FRAME) else 0) | (if (o.NOFRAME == 1) @enumToInt(REDRAW_WINDOW_FLAGS.NOFRAME) else 0));
    }
};
const RDW_INVALIDATE = REDRAW_WINDOW_FLAGS.INVALIDATE;
const RDW_INTERNALPAINT = REDRAW_WINDOW_FLAGS.INTERNALPAINT;
const RDW_ERASE = REDRAW_WINDOW_FLAGS.ERASE;
const RDW_VALIDATE = REDRAW_WINDOW_FLAGS.VALIDATE;
const RDW_NOINTERNALPAINT = REDRAW_WINDOW_FLAGS.NOINTERNALPAINT;
const RDW_NOERASE = REDRAW_WINDOW_FLAGS.NOERASE;
const RDW_NOCHILDREN = REDRAW_WINDOW_FLAGS.NOCHILDREN;
const RDW_ALLCHILDREN = REDRAW_WINDOW_FLAGS.ALLCHILDREN;
const RDW_UPDATENOW = REDRAW_WINDOW_FLAGS.UPDATENOW;
const RDW_ERASENOW = REDRAW_WINDOW_FLAGS.ERASENOW;
const RDW_FRAME = REDRAW_WINDOW_FLAGS.FRAME;
const RDW_NOFRAME = REDRAW_WINDOW_FLAGS.NOFRAME;

const HGDIOBJ = *opaque {};
const HRGN = HGDIOBJ;

extern "USER32" fn RedrawWindow(
    hWnd: ?w.HWND,
    lprcUpdate: ?*const RECT,
    hrgnUpdate: ?HRGN,
    flags: REDRAW_WINDOW_FLAGS,
) callconv(w.WINAPI) BOOL;

// Data exchange
extern "USER32" fn GetClipboardData(
    uFormat: u32,
) callconv(w.WINAPI) ?w.HANDLE;

extern "USER32" fn GetClipboardOwner() callconv(w.WINAPI) ?w.HWND;

extern "USER32" fn OpenClipboard(
    hWndNewOwner: ?w.HWND,
) callconv(w.WINAPI) BOOL;

extern "USER32" fn CloseClipboard() callconv(w.WINAPI) BOOL;

extern "USER32" fn AddClipboardFormatListener(
    hwnd: ?w.HWND,
) callconv(w.WINAPI) BOOL;

// TODO: this type is limited to platform 'windows6.0.6000'
extern "USER32" fn RemoveClipboardFormatListener(
    hwnd: ?w.HWND,
) callconv(w.WINAPI) BOOL;

extern "USER32" fn GetPriorityClipboardFormat(
    paFormatPriorityList: [*]u32,
    cFormats: i32,
) callconv(w.WINAPI) i32;

// System
extern "USER32" fn PostQuitMessage(
    nExitCode: i32,
) callconv(w.WINAPI) void;

const CLIPBOARD_FORMATS = enum(u32) {
    TEXT = 1,
    BITMAP = 2,
    METAFILEPICT = 3,
    SYLK = 4,
    DIF = 5,
    TIFF = 6,
    OEMTEXT = 7,
    DIB = 8,
    PALETTE = 9,
    PENDATA = 10,
    RIFF = 11,
    WAVE = 12,
    UNICODETEXT = 13,
    ENHMETAFILE = 14,
    HDROP = 15,
    LOCALE = 16,
    DIBV5 = 17,
    MAX = 18,
    OWNERDISPLAY = 128,
    DSPTEXT = 129,
    DSPBITMAP = 130,
    DSPMETAFILEPICT = 131,
    DSPENHMETAFILE = 142,
    PRIVATEFIRST = 512,
    PRIVATELAST = 767,
    GDIOBJFIRST = 768,
    GDIOBJLAST = 1023,
};
const CF_TEXT = CLIPBOARD_FORMATS.TEXT;
const CF_BITMAP = CLIPBOARD_FORMATS.BITMAP;
const CF_METAFILEPICT = CLIPBOARD_FORMATS.METAFILEPICT;
const CF_SYLK = CLIPBOARD_FORMATS.SYLK;
const CF_DIF = CLIPBOARD_FORMATS.DIF;
const CF_TIFF = CLIPBOARD_FORMATS.TIFF;
const CF_OEMTEXT = CLIPBOARD_FORMATS.OEMTEXT;
const CF_DIB = CLIPBOARD_FORMATS.DIB;
const CF_PALETTE = CLIPBOARD_FORMATS.PALETTE;
const CF_PENDATA = CLIPBOARD_FORMATS.PENDATA;
const CF_RIFF = CLIPBOARD_FORMATS.RIFF;
const CF_WAVE = CLIPBOARD_FORMATS.WAVE;
const CF_UNICODETEXT = CLIPBOARD_FORMATS.UNICODETEXT;
const CF_ENHMETAFILE = CLIPBOARD_FORMATS.ENHMETAFILE;
const CF_HDROP = CLIPBOARD_FORMATS.HDROP;
const CF_LOCALE = CLIPBOARD_FORMATS.LOCALE;
const CF_DIBV5 = CLIPBOARD_FORMATS.DIBV5;
const CF_MAX = CLIPBOARD_FORMATS.MAX;
const CF_OWNERDISPLAY = CLIPBOARD_FORMATS.OWNERDISPLAY;
const CF_DSPTEXT = CLIPBOARD_FORMATS.DSPTEXT;
const CF_DSPBITMAP = CLIPBOARD_FORMATS.DSPBITMAP;
const CF_DSPMETAFILEPICT = CLIPBOARD_FORMATS.DSPMETAFILEPICT;
const CF_DSPENHMETAFILE = CLIPBOARD_FORMATS.DSPENHMETAFILE;
const CF_PRIVATEFIRST = CLIPBOARD_FORMATS.PRIVATEFIRST;
const CF_PRIVATELAST = CLIPBOARD_FORMATS.PRIVATELAST;
const CF_GDIOBJFIRST = CLIPBOARD_FORMATS.GDIOBJFIRST;
const CF_GDIOBJLAST = CLIPBOARD_FORMATS.GDIOBJLAST;

// memory
extern "KERNEL32" fn GlobalUnlock(
    hMem: isize,
) callconv(w.WINAPI) BOOL;

extern "KERNEL32" fn GlobalLock(
    hMem: isize,
) callconv(w.WINAPI) ?*anyopaque;

const GLOBAL_ALLOC_FLAGS = enum(u32) {
    HND = 66,
    MEM_FIXED = 0,
    MEM_MOVEABLE = 2,
    MEM_ZEROINIT = 64,
    // PTR = 64, this enum value conflicts with MEM_ZEROINIT
    _,
    pub fn initFlags(o: struct {
        HND: u1 = 0,
        MEM_FIXED: u1 = 0,
        MEM_MOVEABLE: u1 = 0,
        MEM_ZEROINIT: u1 = 0,
    }) GLOBAL_ALLOC_FLAGS {
        return @intToEnum(GLOBAL_ALLOC_FLAGS, (if (o.HND == 1) @enumToInt(GLOBAL_ALLOC_FLAGS.HND) else 0) | (if (o.MEM_FIXED == 1) @enumToInt(GLOBAL_ALLOC_FLAGS.MEM_FIXED) else 0) | (if (o.MEM_MOVEABLE == 1) @enumToInt(GLOBAL_ALLOC_FLAGS.MEM_MOVEABLE) else 0) | (if (o.MEM_ZEROINIT == 1) @enumToInt(GLOBAL_ALLOC_FLAGS.MEM_ZEROINIT) else 0));
    }
};
const GHND = GLOBAL_ALLOC_FLAGS.HND;
const GMEM_FIXED = GLOBAL_ALLOC_FLAGS.MEM_FIXED;
const GMEM_MOVEABLE = GLOBAL_ALLOC_FLAGS.MEM_MOVEABLE;
const GMEM_ZEROINIT = GLOBAL_ALLOC_FLAGS.MEM_ZEROINIT;
const GPTR = GLOBAL_ALLOC_FLAGS.MEM_ZEROINIT;

extern "KERNEL32" fn GlobalAlloc(
    uFlags: GLOBAL_ALLOC_FLAGS,
    dwBytes: usize,
) callconv(w.WINAPI) isize;

extern "KERNEL32" fn GlobalFree(
    hMem: isize,
) callconv(w.WINAPI) isize;

// Shell
const Arch = enum { X86, X64, Arm64 };
const arch: Arch = switch (builtin.target.cpu.arch) {
    .i386 => .X86,
    .x86_64 => .X64,
    .arm, .armeb => .Arm64,
    else => @compileError("unable to determine win32 arch"),
};
const NOTIFY_ICON_DATA_FLAGS = enum(u32) {
    MESSAGE = 1,
    ICON = 2,
    TIP = 4,
    STATE = 8,
    INFO = 16,
    GUID = 32,
    REALTIME = 64,
    SHOWTIP = 128,
    _,
    pub fn initFlags(o: struct {
        MESSAGE: u1 = 0,
        ICON: u1 = 0,
        TIP: u1 = 0,
        STATE: u1 = 0,
        INFO: u1 = 0,
        GUID: u1 = 0,
        REALTIME: u1 = 0,
        SHOWTIP: u1 = 0,
    }) NOTIFY_ICON_DATA_FLAGS {
        return @intToEnum(NOTIFY_ICON_DATA_FLAGS, (if (o.MESSAGE == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.MESSAGE) else 0) | (if (o.ICON == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.ICON) else 0) | (if (o.TIP == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.TIP) else 0) | (if (o.STATE == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.STATE) else 0) | (if (o.INFO == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.INFO) else 0) | (if (o.GUID == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.GUID) else 0) | (if (o.REALTIME == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.REALTIME) else 0) | (if (o.SHOWTIP == 1) @enumToInt(NOTIFY_ICON_DATA_FLAGS.SHOWTIP) else 0));
    }
};
const NIF_MESSAGE = NOTIFY_ICON_DATA_FLAGS.MESSAGE;
const NIF_ICON = NOTIFY_ICON_DATA_FLAGS.ICON;
const NIF_TIP = NOTIFY_ICON_DATA_FLAGS.TIP;
const NIF_STATE = NOTIFY_ICON_DATA_FLAGS.STATE;
const NIF_INFO = NOTIFY_ICON_DATA_FLAGS.INFO;
const NIF_GUID = NOTIFY_ICON_DATA_FLAGS.GUID;
const NIF_REALTIME = NOTIFY_ICON_DATA_FLAGS.REALTIME;
const NIF_SHOWTIP = NOTIFY_ICON_DATA_FLAGS.SHOWTIP;
const NOTIFYICONDATAA = switch (arch) {
    .X64, .Arm64 => extern struct {
        cbSize: u32,
        hWnd: ?w.HWND,
        uID: u32,
        uFlags: NOTIFY_ICON_DATA_FLAGS,
        uCallbackMessage: u32,
        hIcon: ?w.HICON,
        szTip: [128]w.CHAR,
        dwState: u32,
        dwStateMask: u32,
        szInfo: [256]w.CHAR,
        Anonymous: extern union {
            uTimeout: u32,
            uVersion: u32,
        },
        szInfoTitle: [64]w.CHAR,
        dwInfoFlags: u32,
        guidItem: Guid,
        hBalloonIcon: ?w.HICON,
    },
    .X86 => packed struct {
        cbSize: u32,
        hWnd: ?w.HWND,
        uID: u32,
        uFlags: NOTIFY_ICON_DATA_FLAGS,
        uCallbackMessage: u32,
        hIcon: ?w.HICON,
        szTip: [128]w.CHAR,
        dwState: u32,
        dwStateMask: u32,
        szInfo: [256]w.CHAR,
        Anonymous: packed union {
            uTimeout: u32,
            uVersion: u32,
        },
        szInfoTitle: [64]w.CHAR,
        dwInfoFlags: u32,
        guidItem: Guid,
        hBalloonIcon: ?w.HICON,
    },
};
const NOTIFY_ICON_MESSAGE = enum(u32) {
    ADD = 0,
    MODIFY = 1,
    DELETE = 2,
    SETFOCUS = 3,
    SETVERSION = 4,
};
const NIM_ADD = NOTIFY_ICON_MESSAGE.ADD;
const NIM_MODIFY = NOTIFY_ICON_MESSAGE.MODIFY;
const NIM_DELETE = NOTIFY_ICON_MESSAGE.DELETE;
const NIM_SETFOCUS = NOTIFY_ICON_MESSAGE.SETFOCUS;
const NIM_SETVERSION = NOTIFY_ICON_MESSAGE.SETVERSION;

const WM_LBUTTONDOWN = @as(u32, 513);
const WM_LBUTTONUP = @as(u32, 514);
const WM_LBUTTONDBLCLK = @as(u32, 515);

extern "SHELL32" fn Shell_NotifyIconA(
    dwMessage: NOTIFY_ICON_MESSAGE,
    lpData: ?*NOTIFYICONDATAA,
) callconv(w.WINAPI) BOOL;

pub fn typedConst(comptime T: type, comptime value: anytype) T {
    return typedConst2(T, T, value);
}

pub fn typedConst2(comptime ReturnType: type, comptime SwitchType: type, comptime value: anytype) ReturnType {
    const target_type_error = @as([]const u8, "typedConst cannot convert to " ++ @typeName(ReturnType));
    const value_type_error = @as([]const u8, "typedConst cannot convert " ++ @typeName(@TypeOf(value)) ++ " to " ++ @typeName(ReturnType));

    switch (@typeInfo(SwitchType)) {
        .Int => |target_type_info| {
            if (value >= std.math.maxInt(SwitchType)) {
                if (target_type_info.signedness == .signed) {
                    const UnsignedT = @Type(std.builtin.TypeInfo{ .Int = .{ .signedness = .unsigned, .bits = target_type_info.bits } });
                    return @bitCast(SwitchType, @as(UnsignedT, value));
                }
            }
            return value;
        },
        .Pointer => |target_type_info| switch (target_type_info.size) {
            .One, .Many, .C => {
                switch (@typeInfo(@TypeOf(value))) {
                    .ComptimeInt, .Int => {
                        const usize_value = if (value >= 0) value else @bitCast(usize, @as(isize, value));
                        return @intToPtr(ReturnType, usize_value);
                    },
                    else => @compileError(value_type_error),
                }
            },
            else => target_type_error,
        },
        .Optional => |target_type_info| switch (@typeInfo(target_type_info.child)) {
            .Pointer => return typedConst2(ReturnType, target_type_info.child, value),
            else => target_type_error,
        },
        .Enum => |_| switch (@typeInfo(@TypeOf(value))) {
            .Int => return @intToEnum(ReturnType, value),
            else => target_type_error,
        },
        else => @compileError(target_type_error),
    }
}

const Guid = extern union {
    Ints: extern struct {
        a: u32,
        b: u16,
        c: u16,
        d: [8]u8,
    },
    Bytes: [16]u8,

    const big_endian_hex_offsets = [16]u6{ 0, 2, 4, 6, 9, 11, 14, 16, 19, 21, 24, 26, 28, 30, 32, 34 };
    const little_endian_hex_offsets = [16]u6{ 6, 4, 2, 0, 11, 9, 16, 14, 19, 21, 24, 26, 28, 30, 32, 34 };
    const hex_offsets = switch (builtin.target.cpu.arch.endian()) {
        .Big => big_endian_hex_offsets,
        .Little => little_endian_hex_offsets,
    };

    pub fn initString(s: []const u8) Guid {
        var guid = Guid{ .Bytes = undefined };
        for (hex_offsets) |hex_offset, i| {
            //guid.Bytes[i] = decodeHexByte(s[offset..offset+2]);
            guid.Bytes[i] = decodeHexByte([2]u8{ s[hex_offset], s[hex_offset + 1] });
        }
        return guid;
    }
};
comptime {
    std.debug.assert(@sizeOf(Guid) == 16);
}

// TODO: is this in the standard lib somewhere?
fn hexVal(c: u8) u4 {
    if (c <= '9') return @intCast(u4, c - '0');
    if (c >= 'a') return @intCast(u4, c + 10 - 'a');
    return @intCast(u4, c + 10 - 'A');
}

// TODO: is this in the standard lib somewhere?
fn decodeHexByte(hex: [2]u8) u8 {
    return @intCast(u8, hexVal(hex[0])) << 4 | hexVal(hex[1]);
}

pub const IDI_APPLICATION = typedConst([*:0]const u16, @as(u32, 32512));

pub extern "USER32" fn LoadIconW(
    hInstance: ?w.HINSTANCE,
    lpIconName: ?[*:0]const u16,
) callconv(w.WINAPI) ?w.HICON;

// resource.h
// #define IDD_MFPLAYBACK_DIALOG           102
// #define IDM_EXIT                        105
// #define IDC_MFPLAYBACK                  109
// #define IDD_OPENURL                     129
// #define IDC_EDIT_URL                    1000
// #define ID_FILE_OPENFILE                32771
// #define ID_FILE_OPENURL                 32772
// #define IDC_STATIC                      -1
const IDM_EXIT = @as(u32, 105);

var h_instance: w.HINSTANCE = undefined;

fn getWinStyleString(comptime buflen: u32, str: []const u8) [buflen]u8 {
    var buf: [buflen]u8 = .{0} ** buflen;
    for (str) |c, i| buf[i] = c;
    return buf;
}

// Globals
var uFormat: w.UINT = @enumToInt(CF_TEXT);
var fAuto: w.BOOL = w.TRUE;
var icon_data: NOTIFYICONDATAA = undefined;
var window_state: union(enum(u8)) {
    INITIAL = 0,
    NORMAL = 1,
    MINIMIZED = 2,
} = .INITIAL;
var gpa: std.heap.GeneralPurposeAllocator(.{}) = undefined;
var contents: [:0]const u8 = "";
var free_contents = false;
var clip: *Clipboard = undefined;

pub export fn wWinMain(hInstance: w.HINSTANCE, hPrevInstance: ?w.HINSTANCE, lpCmdLine: w.PWSTR, nCmdShow: w.INT) w.INT {
    _ = hPrevInstance;
    _ = lpCmdLine;
    _ = nCmdShow;

    gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();

    var clippy = Clipboard.init(gpa.allocator()) catch |e| {
        std.log.err("Could not initialize common clipboard action. {}. Exiting", .{e});
        if (@errorReturnTrace()) |trace| {
            std.debug.dumpStackTrace(trace.*);
        }
        return 1;
    };
    clip = &clippy;
    defer clippy.deinit();

    h_instance = hInstance;

    // Register the window class.
    var wc: WNDCLASSA = .{
        .lpszClassName = "Clipboard watcher",
        .hInstance = hInstance,
        .lpfnWndProc = MainWndProc,
        // default vals
        .style = WNDCLASS_STYLES.initFlags(.{}),
        .cbClsExtra = 0,
        .cbWndExtra = 0,
        .hIcon = null,
        .hCursor = null,
        .hbrBackground = null,
        .lpszMenuName = null,
    };

    _ = RegisterClassA(&wc);

    // Create the window.

    var hwnd = CreateWindowExA(
        WINDOW_EX_STYLE.initFlags(.{}), // Optional window styles.
        wc.lpszClassName.?, // Window class
        "Windows program - in zig", // Window text
        WS_OVERLAPPEDWINDOW, // Window style

        // Size and position
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        CW_USEDEFAULT,

        null, // Parent window: use HWND_MESSAGE to make this completely invisible
        null, // Menu
        hInstance, // Instance handle
        null, // Additional application data
    );

    if (hwnd == null)
        return 0;

    UpdateContents(hwnd.?);
    // Run the message loop.

    var msg: MSG = .{
        .hwnd = null,
        .message = 0,
        .wParam = 0,
        .lParam = 0,
        .time = 0,
        .pt = .{
            .x = 0,
            .y = 0,
        },
    };

    while (GetMessageA(&msg, null, 0, 0) > 0) {
        _ = TranslateMessage(&msg);
        _ = DispatchMessageA(&msg);
    }

    return 0;
}

fn MainWndProc(hwnd: w.HWND, uMsg: u32, wParam: w.WPARAM, lParam: w.LPARAM) callconv(w.WINAPI) w.LRESULT { //APIENTRY {
    const allocator = gpa.allocator();

    switch (uMsg) {
        WM_PAINT => {
            var ps: PAINTSTRUCT = .{
                .hdc = null,
                .fErase = 0,
                .rcPaint = .{
                    .left = 0,
                    .top = 0,
                    .right = 0,
                    .bottom = 0,
                },
                .fRestore = 0,
                .fIncUpdate = 0,
                .rgbReserved = [_]u8{0} ** 32,
            };
            var rc: RECT = .{
                .left = 0,
                .top = 0,
                .right = 0,
                .bottom = 0,
            };
            var hdc = BeginPaint(hwnd, &ps);

            // Branch depending on the clipboard format.
            switch (uFormat) {
                @enumToInt(CF_OWNERDISPLAY) => {
                    // hwndOwner = GetClipboardOwner();
                    // hglb = GlobalAlloc(GMEM_MOVEABLE,
                    //     sizeof(PAINTSTRUCT));
                    // lpps = GlobalLock(hglb);
                    // memcpy(lpps, &ps, sizeof(PAINTSTRUCT));
                    // GlobalUnlock(hglb);
                    //
                    // SendMessage(hwndOwner, WM_PAINTCLIPBOARD,
                    //     (WPARAM) hwnd, (LPARAM) hglb);
                    //
                    // GlobalFree(hglb);
                },

                @enumToInt(CF_BITMAP) => {
                    // hdcMem = CreateCompatibleDC(hdc);
                    // if (hdcMem != NULL)
                    // {
                    //     if (OpenClipboard(hwnd))
                    //     {
                    //         hbm = (HBITMAP)
                    //             GetClipboardData(uFormat);
                    //         SelectObject(hdcMem, hbm);
                    //         GetClientRect(hwnd, &rc);
                    //
                    //         BitBlt(hdc, 0, 0, rc.right, rc.bottom,
                    //             hdcMem, 0, 0, SRCCOPY);
                    //         CloseClipboard();
                    //     }
                    //     DeleteDC(hdcMem);
                    // }
                },
                @enumToInt(CF_TEXT) => {
                    _ = GetClientRect(hwnd, &rc);
                    _ = DrawTextA(hdc, contents, -1, &rc, DT_LEFT);
                },

                @enumToInt(CF_ENHMETAFILE) => {
                    // if (OpenClipboard(hwnd))
                    // {
                    //     hemf = GetClipboardData(uFormat);
                    //     GetClientRect(hwnd, &rc);
                    //     PlayEnhMetaFile(hdc, hemf, &rc);
                    //     CloseClipboard();
                    // }
                },
                0 => {
                    _ = GetClientRect(hwnd, &rc);
                    _ = DrawTextA(
                        hdc,
                        "The clipboard is empty.",
                        -1,
                        &rc,
                        @intToEnum(DRAW_TEXT_FORMAT, @enumToInt(DT_CENTER) | @enumToInt(DT_SINGLELINE) |
                            @enumToInt(DT_VCENTER)),
                    );
                },

                else => {
                    _ = GetClientRect(hwnd, &rc);
                    _ = DrawTextA(
                        hdc,
                        "Unable to display format.",
                        -1,
                        &rc,
                        @intToEnum(DRAW_TEXT_FORMAT, @enumToInt(DT_CENTER) | @enumToInt(DT_SINGLELINE) |
                            @enumToInt(DT_VCENTER)),
                    );
                },
            }
            _ = EndPaint(hwnd, &ps);
        },
        WM_SIZE => {
            if (uFormat == @enumToInt(CF_OWNERDISPLAY)) {
                // I don't believe this will ever execute and can be removed
                var hwndOwner = GetClipboardOwner().?;
                var hglb = GlobalAlloc(GMEM_MOVEABLE, @sizeOf(RECT));
                defer _ = GlobalFree(hglb);
                var lprc = GlobalLock(hglb);
                _ = GetClientRect(hwnd, @ptrCast(?*RECT, @alignCast(@alignOf(?*RECT), lprc)));
                _ = GlobalUnlock(hglb);

                // hwnd and hglb are just casts
                _ = SendMessageA(hwndOwner, WM_SIZECLIPBOARD, @ptrToInt(hwnd), hglb);
            }
        },
        WM_CLIPBOARDUPDATE => { // Message arrives only after AddClipboardFormatListener
            UpdateContents(hwnd);
            if (RedrawWindow(hwnd, null, null, RDW_INVALIDATE) != w.TRUE) {
                _ = MessageBoxA(hwnd, "Redraw failed", "Error", 0);
            }
        },
        WM_CREATE => {
            if (AddClipboardFormatListener(hwnd) != w.TRUE) {
                _ = MessageBoxA(hwnd, "Could not add clipboard listener", "Error", 0);
                return -1; // oops
            }
            var tip = getWinStyleString(128, "Clipboard processor");
            icon_data = .{
                .cbSize = @sizeOf(@TypeOf(icon_data)),
                .hWnd = hwnd,
                .uFlags = NOTIFY_ICON_DATA_FLAGS.initFlags(.{
                    .MESSAGE = 1,
                    .ICON = 1,
                    .TIP = 1,
                }),
                .uCallbackMessage = WM_TRAY,
                .uID = 4242,
                .hIcon = LoadIconW(null, IDI_APPLICATION), // TODO: Custom icon
                .szTip = tip,
                .dwState = 0,
                .dwStateMask = 0,
                .szInfo = getWinStyleString(256, "Info"),
                .Anonymous = .{ .uTimeout = 0 },
                .szInfoTitle = getWinStyleString(64, "Info Title"),
                .dwInfoFlags = 0,
                .guidItem = Guid.initString("3e781b84-3ffd-44a1-b3ab-11d0f90136f9"), // generated from duckduckgo
                .hBalloonIcon = null,
            };

            if (Shell_NotifyIconA(NIM_ADD, &icon_data) != w.TRUE) {
                _ = MessageBoxA(hwnd, "Notification Icon failed to create", "Error", 0);
                return -1; // oops
            }
        },

        WM_TRAY => {
            switch (lParam) {
                WM_LBUTTONUP => {
                    switch (window_state) {
                        .INITIAL => {
                            _ = ShowWindow(hwnd, SHOW_WINDOW_CMD.SHOWNORMAL);
                            window_state = .NORMAL;
                        },
                        .NORMAL => {
                            _ = ShowWindow(hwnd, SHOW_WINDOW_CMD.HIDE);
                            window_state = .MINIMIZED;
                        },
                        .MINIMIZED => {
                            _ = ShowWindow(hwnd, SHOW_WINDOW_CMD.SHOW);
                            window_state = .NORMAL;
                        },
                    }
                },
                else => {},
            }
        },
        WM_DESTROY => {
            // remove last item before we go down
            if (free_contents) {
                allocator.free(contents);
            }
            _ = RemoveClipboardFormatListener(hwnd);
            _ = Shell_NotifyIconA(NIM_DELETE, &icon_data);
            PostQuitMessage(0);
        },

        WM_COMMAND => {
            switch (wParam & 0xffffffff) {
                IDM_EXIT => _ = DestroyWindow(hwnd),
                else => {
                    fAuto = w.FALSE;
                    uFormat = @intCast(w.UINT, wParam & 0xffffffff); // LOWORD(wParam);
                    _ = InvalidateRect(hwnd, null, w.TRUE);
                },
            }
        },
        else => return DefWindowProcA(hwnd, uMsg, wParam, lParam),
    }
    return 0;
}

fn UpdateContents(hwnd: w.HWND) void {
    const allocator = gpa.allocator();
    // Get a lock on the clipboard
    if (OpenClipboard(hwnd) > 0) {
        defer _ = CloseClipboard();
        var hglb = GetClipboardData(uFormat).?;
        const hnd = @intCast(isize, @ptrToInt(hglb));
        var lpstr = GlobalLock(hnd).?;
        defer _ = GlobalUnlock(hnd);

        if (free_contents) {
            allocator.free(contents);
        }
        free_contents = true;
        contents = std.fmt.allocPrintZ(allocator, "{s}", .{@ptrCast([*:0]const u8, lpstr)}) catch blk: {
            _ = MessageBoxA(hwnd, "Allocation failed copying clipboard contents", "Error", 0);
            free_contents = false;
            break :blk "";
        };
    }
    _ = async clip.clipboardChanged(contents); // Notify common handler
}
